"""Context class used for reading commands in various contexts."""
from collections import OrderedDict
from typing import Any, Dict, Iterable, List, Optional
from typing import OrderedDict as OrderedDictType

from statue.constants import ALIASES, ALLOWED_BY_DEFAULT, HELP, PARENT
from statue.exceptions import ContextCircularParentingError


class Context:
    """
    Class representing a command context.

    Commands can be run in different contexts. Contexts allow you to customize the
    command arguments according to the context you are using. For ex
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str,
        help: str,  # pylint: disable=redefined-builtin
        aliases: Optional[Iterable[str]] = None,
        parent: Optional["Context"] = None,
        allowed_by_default: bool = False,
    ):
        """
        Context constructor.

        :param name: Name of the context
        :type name: str
        :param help: Short help string to describe context
        :type help: str
        :param aliases: List of possible aliases of the context
        :type aliases: Optional[Iterable[str]]
        :param parent: Optional parent context for this context
        :type parent: Optional[Context]
        :param allowed_by_default: Allow this context for all commands by default
        :type allowed_by_default: bool
        """
        self.name = name
        self.help = help
        self.aliases = list(aliases) if aliases is not None else []
        self.parent = parent
        self.allowed_by_default = allowed_by_default

    @property
    def parent(self) -> Optional["Context"]:
        """Get parent of this context."""
        return self._parent

    @parent.setter
    def parent(self, parent: Optional["Context"]):
        """
        Set parent of this context.

        :param parent: Desired parent
        :type parent: Context
        :raises ContextCircularParentingError: Raised when trying to set a parent
            which is a child of this context.
        """
        if parent is not None and parent.is_child_of(self):
            raise ContextCircularParentingError(self.name, parent.name)
        self._parent = parent

    @property
    def parents(self) -> List["Context"]:
        """Get all parents recursively for this context."""
        if self.parent is None:
            return []
        return [self.parent, *self.parent.parents]

    def __eq__(self, other: object) -> bool:
        """
        Check equality between two contexts.

        :param other: other object to compare to
        :type other: object
        :return: are contexts equal
        :rtype: bool
        """
        return (
            isinstance(other, Context)
            and self.name == other.name
            and self.help == other.help
            and self.aliases == other.aliases
            and self.parent == other.parent
            and self.allowed_by_default == other.allowed_by_default
        )

    def __repr__(self) -> str:
        """
        String representation of the context.

        :return: context as string
        :rtype: str
        """
        return (
            "Context("
            f"name='{self.name}', "
            f"help='{self.help}', "
            f"aliases={self.aliases}, "
            f"parent={self.parent}, "
            f"allowed_by_default={self.allowed_by_default}"
            ")"
        )

    def __hash__(self) -> int:
        """
        Hash context according to its name.

        :return: hash
        :rtype: int
        """
        return hash(self.name)

    @property
    def all_names(self) -> List[str]:
        """List of all possible names."""
        return [self.name, *self.aliases]

    def clear_aliases(self):
        """Remove all aliases of context."""
        self.aliases.clear()

    def is_child_of(self, parent: "Context") -> bool:
        """
        Checks if this context is a child of another context.

        :param parent: Possible parent for this context.
        :type parent: Context
        :return: Is this context child of the given context
        :rtype: bool
        """
        if self.parent == parent:
            return True
        if self.parent is None:
            return False
        return self.parent.is_child_of(parent)

    def is_matching(self, name: str) -> bool:
        """
        Check if a given name is identical to one of the contexts names.

        :param name: Name to be checked
        :type name: str
        :return: Is name equal to context name or one of its aliases
        :rtype: bool
        """
        return name in self.all_names

    def is_matching_recursively(self, name: str) -> bool:
        """
        Check if given name matches this context or its parent.

        :param name: Name to be checked
        :type name: str
        :return: Is name matching to this context or its parent.
        :rtype: bool
        """
        if self.is_matching(name):
            return True
        for parent in self.parents:
            if parent.is_matching(name):
                return True
        return False

    def search_context_instructions(
        self, setups: Dict[str, Any]
    ) -> Optional[Dict[str, Any]]:
        """
        Search for context in setup dictionary.

        :param setups: Setup to look for the context in
        :type setups: Dict[str, Any]
        :return: Specific setups with context
        :rtype: None or Dict[str, Any]
        """
        for name in self.all_names:
            name_setups = setups.get(name, None)
            if name_setups is not None:
                return name_setups
        for parent in self.parents:
            instructions = parent.search_context_instructions(setups)
            if instructions is not None:
                return instructions
        return None

    def as_dict(self) -> OrderedDictType[str, Any]:
        """
        Encode context as a dictionary.

        This is used in order to serialize the context in a configuration file.

        :return: Serialized representation dictionary
        :rtype: OrderedDict[str, Any]
        """
        context_dict: OrderedDict[str, Any] = OrderedDict()
        context_dict[HELP] = self.help
        if len(self.aliases) != 0:
            context_dict[ALIASES] = self.aliases
        if self.parent is not None:
            context_dict[PARENT] = self.parent.name
        if self.allowed_by_default:
            context_dict[ALLOWED_BY_DEFAULT] = True
        return context_dict



"""Exceptions module."""
from typing import List, Optional


class StatueException(Exception):
    """Exceptions base for Statue."""


# Configuration related exceptions


class StatueConfigurationError(StatueException):
    """Exception regarding Statue configuration."""

    def __init__(self, message: str, location: Optional[List[str]] = None):
        """
        Constructor.

        :param message: Message of the error
        :type message: str
        :param location: Optional. Where the exception was thrown in the configuration
        :type location: Optional[List[str]]
        """
        super().__init__(message)
        self.location = [] if location is None else location

    def __str__(self) -> str:
        """
        String representation method.

        :return: String representation of the exception
        :rtype: str
        """
        message = super().__str__()
        if len(self.location) == 0:
            return message
        return f"{message} ({' -> '.join(self.location)})"

    def append_location_item(self, location: str):
        """
        Add high-hierarchy location specifier.

        :param location: Location specifier to be added
        :type location: str
        """
        self.location.insert(0, location)


class MissingConfiguration(StatueConfigurationError):
    """Part of the Statue configuration is missing."""

    message = "Statue was unable to load configuration"

    def __init__(self):
        """Exception constructor."""
        super().__init__(self.message)


class InvalidConfiguration(StatueConfigurationError):
    """Some of statue's configurations are invalid."""


class MissingHelpString(InvalidConfiguration):
    """Help string is missing from configuration."""

    message = "help string is missing"

    def __init__(self, location: Optional[List[str]] = None):
        """
        Constructor.

        :param location: Optional. Where the exception was thrown in the configuration
        :type location: Optional[List[str]]
        """
        super().__init__(message=self.message, location=location)


class InconsistentConfiguration(StatueConfigurationError):
    """Some of statue's configurations are inconsistent."""


# Command related exceptions


class UnknownCommand(StatueException):
    """Command isn't recognized."""

    def __init__(self, command_name: str) -> None:
        """
        Exception constructor.

        :param command_name: Name of the unfound command
        :type command_name: str
        """
        super().__init__(f'Could not find command named "{command_name}"')


class InvalidCommand(StatueException):
    """Command doesn't fit restrictions."""


class CommandsMapBuilderError(StatueException):
    """Could not build commands map."""


class CommandExecutionError(StatueException):
    """Command cannot be executed."""

    def __init__(self, command_name: str) -> None:
        """
        Exception constructor.

        :param command_name: Command name
        :type command_name: str
        """
        super().__init__(
            f'Cannot execute "{command_name}" because it is not installed.'
        )


# Context related exceptions


class UnknownContext(StatueException):
    """Context isn't recognized."""

    def __init__(self, context_name: str) -> None:
        """
        Exception constructor.

        :param context_name: Name of the unfound context
        :type context_name: str
        """
        super().__init__(f'Could not find context named "{context_name}"')
        self.context_name = context_name


class ContextCircularParentingError(StatueException):
    """Parent cannot be a child of its child."""

    def __init__(self, context1: str, context2: str):
        """
        Exception constructor.

        :param context1: Name of first context
        :type context1: str
        :param context2: Name of second context
        :type context2: str
        """
        super().__init__(
            "Cannot set circular parenting between " f'"{context1}" and "{context2}"'
        )


# Template related exceptions


class StatueTemplateError(StatueException):
    """Exception regarding Statue template."""


class UnknownTemplate(StatueTemplateError):
    """Template isn't recognized."""

    def __init__(self, template_name: str) -> None:
        """
        Exception constructor.

        :param template_name: Name of the unfound template
        :type template_name: str
        """
        super().__init__(f'Could not find template named "{template_name}"')


# Cache related exceptions


class CacheError(StatueException):
    """Cache related exception."""



"""Get Statue global configuration."""
import sys
from collections import OrderedDict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, FrozenSet, List, Optional
from typing import OrderedDict as OrderedDictType
from typing import TypeVar, Union

import tomli
import tomli_w

from statue.cache import Cache
from statue.command import Command
from statue.command_builder import CommandBuilder
from statue.commands_filter import CommandsFilter
from statue.commands_map import CommandsMap
from statue.config.commands_repository import CommandsRepository
from statue.config.contexts_repository import ContextsRepository
from statue.config.sources_repository import SourcesRepository
from statue.constants import (
    CACHE,
    COMMANDS,
    CONTEXTS,
    DEFAULT_HISTORY_SIZE,
    GENERAL,
    HISTORY_SIZE,
    MODE,
    SOURCES,
)
from statue.context import Context
from statue.exceptions import (
    InvalidConfiguration,
    MissingConfiguration,
    StatueConfigurationError,
)
from statue.runner import RunnerMode

if sys.version_info < (3, 9):  # pragma: no cover
    from importlib_resources.abc import Traversable
else:  # pragma: no cover
    from importlib.abc import Traversable


T = TypeVar("T")


@dataclass
class Configuration:
    """Configuration singleton for statue."""

    cache: Cache
    default_mode: RunnerMode = field(default=RunnerMode.DEFAULT_MODE)
    contexts_repository: ContextsRepository = field(default_factory=ContextsRepository)
    commands_repository: CommandsRepository = field(default_factory=CommandsRepository)
    sources_repository: SourcesRepository = field(default_factory=SourcesRepository)

    def remove_context(self, context: Context):
        """
        Remove all references of a context from the configuration.

        :param context: Context object to be removed
        :type context: Context
        """
        for source in self.sources_repository.sources_list:
            original_filter = self.sources_repository[source]
            self.sources_repository[source] = CommandsFilter(
                contexts=[
                    filter_context
                    for filter_context in original_filter.contexts
                    if filter_context != context
                ],
                allowed_commands=original_filter.allowed_commands,
                denied_commands=original_filter.denied_commands,
            )
        for command_builder in self.commands_repository:
            command_builder.remove_context(context)
        self.contexts_repository.remove_context(context)

    def remove_command(self, command_builder: CommandBuilder):
        """
        Remove all references of a command builder from the configuration.

        :param command_builder: Command builder to be removed
        :type command_builder: CommandBuilder
        """
        for source in self.sources_repository.sources_list:
            original_filter = self.sources_repository[source]
            self.sources_repository[source] = CommandsFilter(
                contexts=original_filter.contexts,
                allowed_commands=self._none_or_remove(
                    original_filter.allowed_commands, command_builder.name
                ),
                denied_commands=self._none_or_remove(
                    original_filter.denied_commands, command_builder.name
                ),
            )
        self.commands_repository.remove_command_builder(command_builder)

    def build_commands_map(
        self, sources: List[Path], commands_filter: CommandsFilter
    ) -> CommandsMap:
        """
        Build commands map from sources list and a commands filter.

        :param sources: Sources list of the commands map
        :type sources: List[Path]
        :param commands_filter: Base filter to choose commands with
        :type commands_filter: CommandsFilter
        :return: Commands map with sources as keys
        :rtype: CommandsMap
        """
        commands_map = CommandsMap()
        for source in sources:
            commands = self.build_commands(
                CommandsFilter.merge(commands_filter, self.sources_repository[source])
            )
            if len(commands) != 0:
                commands_map[source] = commands
        return commands_map

    def build_commands(self, commands_filter: CommandsFilter) -> List[Command]:
        """
        Read commands with given constraints.

        :param commands_filter: Filter to choose commands according to.
        :type commands_filter: CommandsFilter
        :return: List of commands according to constraints
        :rtype: List[Command]
        """
        return [
            command_builder.build_command(*commands_filter.contexts)
            for command_builder in self.commands_repository
            if commands_filter.pass_filter(command_builder)
        ]

    def as_dict(self) -> OrderedDictType[str, Any]:
        """
        Encode configuration as a dictionary.

        This is used in order to serialize the configuration to be later saved
        in a file.

        :return: Serialized representation dictionary
        :rtype: OrderedDict[str, Any]
        """
        general_dict = OrderedDict(
            [
                (MODE, self.default_mode.name.lower()),
                (HISTORY_SIZE, self.cache.history_size),
            ]
        )
        if not self.cache.enabled:
            general_dict[CACHE] = False
        return OrderedDict(
            [
                (GENERAL, general_dict),
                (CONTEXTS, self.contexts_repository.as_dict()),
                (COMMANDS, self.commands_repository.as_dict()),
                (SOURCES, self.sources_repository.as_dict()),
            ]
        )

    def to_toml(self, path: Path):
        """
        Save configuration to toml file.

        :param path: Path to save configuration in
        :type path: Path
        """
        with path.open(mode="wb") as configuration_file:
            tomli_w.dump(self.as_dict(), configuration_file)

    @classmethod
    def from_file(
        cls,
        config_path: Optional[Union[Path, Traversable]] = None,
        cache_dir: Optional[Path] = None,
    ) -> "Configuration":
        """
        Load statue configuration.

        This method combines default configuration with user-defined configuration, read
        from configuration file.

        :param config_path: User-defined file path containing
            repository-specific configurations
        :type config_path: Optional[Path]
        :param cache_dir: Optional Caching directory
        :type cache_dir: Optional[Path]
        :return: Configuration instance
        :rtype: Configuration
        :raises MissingConfiguration: Raised when could not load
        """
        if config_path is None:
            config_path = cls.configuration_path()
        if isinstance(config_path, Path) and not config_path.exists():
            raise MissingConfiguration()
        with config_path.open(mode="rb") as configuration_file:
            statue_config = tomli.load(configuration_file)
        cache_dir = cls.cache_path(Path.cwd()) if cache_dir is None else cache_dir
        return cls.from_dict(cache_dir=cache_dir, statue_config_dict=statue_config)

    @classmethod
    def from_dict(
        cls, cache_dir: Path, statue_config_dict: Dict[str, Any]
    ) -> "Configuration":
        """
        Build configuration from a loaded config map.

        :param cache_dir: Directory for keeping cache.
        :type cache_dir: Path
        :param statue_config_dict: Configuration map as loaded from config file
        :type statue_config_dict: Dict[str, Any]
        :return: Built configuration instance
        :type: Configuration
        :raises InvalidConfiguration: Raised when some fields are invalid
            in configuration
        """
        general_configuration = statue_config_dict.get(GENERAL, {})
        history_size = general_configuration.get(HISTORY_SIZE, DEFAULT_HISTORY_SIZE)
        cached_enabled = general_configuration.get(CACHE, True)
        cache = Cache(
            cache_root_directory=cache_dir, size=history_size, enabled=cached_enabled
        )
        mode = RunnerMode.DEFAULT_MODE
        if MODE in general_configuration:
            mode_string = general_configuration[MODE].upper()
            try:
                mode = RunnerMode[mode_string]
            except KeyError as error:
                raise InvalidConfiguration(
                    f"Got unexpected runner mode {mode_string}", location=[GENERAL]
                ) from error
        contexts_repository = cls.build_contexts_repository(statue_config_dict)
        commands_repository = cls.build_commands_repository(
            statue_config_dict, contexts_repository
        )
        sources_repository = SourcesRepository.from_dict(
            config=statue_config_dict.get(SOURCES, {}),
            contexts_repository=contexts_repository,
        )
        return Configuration(
            cache=cache,
            default_mode=mode,
            contexts_repository=contexts_repository,
            commands_repository=commands_repository,
            sources_repository=sources_repository,
        )

    @classmethod
    def build_contexts_repository(
        cls, statue_config_dict: Dict[str, Any]
    ) -> ContextsRepository:
        """
        Build contexts repository from configuration dictionary.

        :param statue_config_dict: Configuration map as loaded from config file
        :type statue_config_dict: Dict[str, Any]
        :return: Built contexts repository
        :rtype: ContextsRepository
        :raises StatueConfigurationError: General configuration-related error.
        """
        try:
            return ContextsRepository.from_dict(statue_config_dict.get(CONTEXTS, {}))
        except StatueConfigurationError as error:
            error.append_location_item(CONTEXTS)
            raise error

    @classmethod
    def build_commands_repository(
        cls, statue_config_dict: Dict[str, Any], contexts_repository: ContextsRepository
    ) -> CommandsRepository:
        """
        Build contexts repository from configuration dictionary.

        :param statue_config_dict: Configuration map as loaded from config file
        :type statue_config_dict: Dict[str, Any]
        :param contexts_repository: Contexts repository to get contexts from
        :type contexts_repository: ContextsRepository
        :return: Built contexts repository
        :rtype: CommandsRepository
        :raises StatueConfigurationError: General configuration-related error.
        """
        try:
            return CommandsRepository.from_dict(
                config=statue_config_dict.get(COMMANDS, {}),
                contexts_repository=contexts_repository,
            )
        except StatueConfigurationError as error:
            error.append_location_item(COMMANDS)
            raise error

    @classmethod
    def configuration_path(cls, directory: Optional[Path] = None) -> Path:
        """
        Search for configuration file in directory.

        :param directory: Directory in which the configuration path is supposed to be
        :type directory: Path
        :return: Configuration path location
        :rtype: Path
        """
        if directory is None:
            directory = Path.cwd()
        return directory / "statue.toml"

    @classmethod
    def cache_path(cls, directory: Path) -> Path:
        """
        Default caching directory for statue history saving.

        :param directory: Directory in which cache directory will be created
        :type directory: Path
        :return: Cache directory
        :rtype: Path
        """
        return directory / ".statue"

    @classmethod
    def empty_configuration(cls) -> "Configuration":
        """
        Creates an empty configuration.

        :return: Empty configuration
        :rtype: Configuration
        """
        cache = Cache(size=DEFAULT_HISTORY_SIZE)
        return Configuration(cache=cache)

    @classmethod
    def _none_or_remove(
        cls, optional_set: Optional[FrozenSet[T]], removed_item: T
    ) -> Optional[FrozenSet[T]]:
        if optional_set is None:
            return None
        return frozenset(item for item in optional_set if item != removed_item)



"""Specific arguments manipulations for given context."""
from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional
from typing import OrderedDict as OrderedDictType

from statue.constants import ADD_ARGS, ARGS, CLEAR_ARGS
from statue.exceptions import InconsistentConfiguration


@dataclass
class ContextSpecification:
    """Specific instructions for building command in context."""

    args: Optional[List[str]] = field(default=None)
    add_args: Optional[List[str]] = field(default=None)
    clear_args: bool = field(default=False)

    def __post_init__(self):
        """Validate after initializing."""
        self._validate()

    def update_args(self, args: List[str]) -> List[str]:
        """
        Update command arguments according to instructions.

        :param args: Original arguments
        :type args: List[str]
        :return: Updated arguments
        :rtype: List[str]
        """
        if self.args is not None:
            return self.args
        if self.add_args is not None:
            return args + self.add_args
        if self.clear_args:
            return []
        return args

    def as_dict(self) -> OrderedDictType[str, Any]:
        """
        Encode context specification as a dictionary.

        This is used in order to serialize the context specification in
        a configuration file.

        :return: Serialized representation dictionary
        :rtype: OrderedDict[str, Any]
        """
        specification_as_dict: OrderedDictType[str, Any] = OrderedDict()
        if self.args is not None:
            specification_as_dict[ARGS] = self.args
        if self.add_args is not None:
            specification_as_dict[ADD_ARGS] = self.add_args
        if self.clear_args:
            specification_as_dict[CLEAR_ARGS] = True
        return specification_as_dict

    def _validate(self):
        """
        Validate that the context specification does contradict itself.

        :raises InconsistentConfiguration: raised when context
            specification is inconsistent.
        """
        if self.clear_args and self.args is not None:
            raise InconsistentConfiguration(
                "args and clear_args cannot be both set at the same time",
                location=["args/clear_args"],
            )

        if self.clear_args and self.add_args is not None:
            raise InconsistentConfiguration(
                "add_args and clear_args cannot be both set at the same time",
                location=["add_args/clear_args"],
            )

        if self.args is not None and self.add_args is not None:
            raise InconsistentConfiguration(
                "args and add_args cannot be both set at the same time",
                location=["args/add_args"],
            )

    @classmethod
    def from_dict(
        cls, context_specification_setups: Dict[str, Any]
    ) -> "ContextSpecification":
        """
        Read Context specification from json.

        :param context_specification_setups: Context specification json
        :type context_specification_setups: Dict[str, Any]
        :return: Built context specification
        :rtype: ContextSpecification
        """
        return ContextSpecification(
            args=context_specification_setups.get(ARGS, None),
            add_args=context_specification_setups.get(ADD_ARGS, None),
            clear_args=context_specification_setups.get(CLEAR_ARGS, False),
        )



"""Transforming plane strings into styled ones."""
import click


def source_style(source: str) -> str:
    """
    Styling function to emphasise sources paths.

    :param source: The source name to style
    :type source: str
    :return: Styles source
    :rtype: str
    """
    return click.style(source, fg="cyan")


def name_style(name: str) -> str:
    """
    Styling function to emphasise names.

    :param name: The name to style
    :type name: str
    :return: Styled name
    :rtype: str
    """
    return click.style(name, fg="magenta")


def bullet_style(bullet_name: str) -> str:
    """
    Styling function to emphasise bullet names.

    :param bullet_name: The name to style
    :type bullet_name: str
    :return: Styled bullet name
    :rtype: str
    """
    return click.style(bullet_name, fg="yellow")


def success_style(success_string: str) -> str:
    """
    Styling function to emphasise bullet names.

    :param success_string: The string to style
    :type success_string: str
    :return: Styled success string
    :rtype: str
    """
    return click.style(success_string, fg="green")


def failure_style(failure_string: str) -> str:
    """
    Styling function to emphasise bullet names.

    :param failure_string: The string to style
    :type failure_string: str
    :return: Styled failure string
    :rtype: str
    """
    return click.style(failure_string, fg="red")
